;; Generated package description from qrencode.el  -*- no-byte-compile: t; lexical-binding:t -*-
(define-package "qrencode" "1.5beta4.0.20260906.0" "QRCode encoder" '((emacs "25.1")) :commit "a63a44020f48cd178c762a30104d09f0cfa584fa" :authors '(("Rüdiger Sonderfeld" . "ruediger@c-plusplus.net")) :maintainer '("Rüdiger Sonderfeld" . "ruediger@c-plusplus.net") :keywords '("qrcode" "comm") :url "https://github.com/ruediger/qrencode-el")
